#!/usr/bin/env python
#
######################################
# With this code Figures 5 to 9 of the following manuscript were produced:
#     Grathwohl, Kunstmann, Quinto, Rieder:
#     Approximate inverse for the common offset acquisition geometry in 2D seismic imaging
#     http://www.waves.kit.edu/downloads/CRC1173_Preprint_2016-37.pdf
######################################

#import matplotlib
#matplotlib.use("Agg")

from EllipticRadon2D import *
import numpy as np
import matplotlib.pyplot as plt
from math import floor, ceil
from concurrent import futures


alpha = 2.0 # common offset
gamma = 0.3 # regularization parameter: radius of the mollifier's support
delta = 0.03 # parameter for artifical random noise 

visualization_data = False
visualization_recon = True
normal_noise = False  # normal noise with pointwise relative standard deviation delta
uniform_noise = False # uniform noise with relative strenght in norm delta.
                      # see above cited mamuscript for details

colormap = 'gray' # 'seismic', 'jet'
linecolor = 'black'

p1 = 0.;  q1 = 4.; r1 = 2.0  #Ball 1
p2 = 0.;  q2 = 4.; r2 = 1.0   #Ball 2
p3 = 3.;  q3 = 5.; r3 = 1.5 


tmin =5 #2.*alpha+0.1  #[tmin,tmax] range of diameters (tmin > 2*alpha)
tmax = 20+tmin

Nt = 200  # no. of equi-distant sampling points in [tmin,tmax]

smin = -10. #[smin,smax] range of mid points
smax = 10.

Ns = 200  # no. of equi-distant sampling points in [smin,smax]

xmin = -2.5 # [xmin,xmax] range for reconstrution in x-Variable
xmax = 5.

ymin = 1.5 # [ymin,ymax] range for reconstrution in y-Variable
ymax = 7.

Nx = 100 # Nx number of equi-distant reconstruction points in [xmin,xmax]
Ny = Nx # Ny number of equi-distant reconstruction points in [ymin,ymax] 

t = np.linspace(tmin, tmax, Nt)  #sample points in [tmin,tmax]

s = np.linspace(smin, smax, Ns)  #sample points in [smin,smax]

dt = t[1]-t[0]
ds = s[1]-s[0]

F = np.zeros((Ns,Nt)) 

for i in xrange(Ns):
    sc = cut_off_s(s[i],smax-1,smax)
    for j in xrange(Nt):
        F[i,j] = (calculate_F_Ball_2D(s[i],t[j],alpha,p1,q1,r1) \
      - calculate_F_Ball_2D(s[i],t[j],alpha,p2,q2,r2) + calculate_F_Ball_2D(s[i],t[j],alpha,p3,q3,r3) \
         +calculate_F_floor_2D(s[i],t[j],alpha,6.5)
        )*cut_off_t(t[j],tmax-1,tmax,0.01)*sc

#file_name ='Data_Balls3_floor_alpha%4.2f_s%4.2f_t%4.2f_Ns%3d_Nt%3d' % (alpha,smax,tmax,Ns,Nt)
#np.save(file_name, F)
    
if uniform_noise:
   np.random.seed(53248989)
   noise = 2.*np.random.rand(Ns,Nt)-1.
   norm=0;
   normn=0;
   for i in xrange(Ns):
       for j in xrange(Nt):
           norm=norm+F[i,j]**2*t[j]**2
           normn=normn+noise[i,j]**2*t[j]**2    
   norm=np.sqrt(norm*dt*ds)
   normn=np.sqrt(normn*dt*ds)
   F = F + delta*noise*norm/normn
   
if normal_noise:
   np.random.seed(53248989)
   noise = np.random.randn(Ns,Nt)
   F = F + delta*F*noise

      
if visualization_data:      
   plt.close(1)  
   plt.figure(1)
   imgplot=plt.imshow(np.rot90(F,1),origin='lower', extent=(smin, smax, tmax, tmin),interpolation='nearest') #clim=(-.09, 0.09))
   plt.xlabel(r'midpoint $s$', fontsize=18)
   plt.ylabel(r'travel time (diameter) $t$', fontsize=18)
   plt.title(r'Elliptic 2D-Radon transform (sinogram), $\alpha = $ %4.2f' % alpha)
   plt.colorbar()
   plt.show(1)
   plt.pause(0.0001)
 
x = np.linspace(xmin,xmax,Nx)
y = np.linspace(ymin,ymax,Ny)

f_rec = np.zeros((Nx,Ny))
f_rec2 = np.zeros((Nx,Ny))

def reconstructor(args):
    x1,x2,s,t,data,alpha,gamma = args
    result = 0.
    for i in xrange(len(s)):
        (Tmin,Tmax,dummy1,dummy2)=calculate_T_min_max(x1,x2,s[i], alpha, gamma)
        jmin = max(int(ceil((Tmin-t[0])/(t[1]-t[0])))-1,0)
        jmax = min(int(floor((Tmax-t[0])/(t[1]-t[0])))+1,len(t))
        for j in xrange(jmin,jmax):
            kern = Kernel_e3_2D_Laplace_approx(s[i],t[j],alpha,x1,x2,gamma,Tmin,Tmax)*data[i,j]
            result = result + kern

    return result

liste = [(x[0],y[0],s,t,F,alpha,gamma)]

for i in xrange(Nx):
    for j in xrange(Ny):
        if i+j > 0:
           liste += [(x[i],y[j],s,t,F,alpha,gamma)]
           
           
with futures.ProcessPoolExecutor() as e:
    f = list(e.map(reconstructor,liste))
    
for i in xrange(Nx):
    for j in xrange(Ny):
        f_rec[i,j] = dt*ds*f[j+Ny*i]

if uniform_noise:
   file_name ='Balls3_floor_gam%4.2f_alpha%4.2f_s%4.2f_t%4.2f_Ns%3d_Nt%3d_Nx%2d_uninoise%4.2f' % (gamma,alpha,smax,tmax,Ns,Nt,Nx,delta)
elif normal_noise:
   file_name ='Balls3_floor_gam%4.2f_alpha%4.2f_s%4.2f_t%4.2f_Ns%3d_Nt%3d_Nx%2d_normalnoise%4.2f' % (gamma,alpha,smax,tmax,Ns,Nt,Nx,delta)
else:
   file_name ='Balls3_floor_gam%4.2f_alpha%4.2f_s%4.2f_t%4.2f_Ns%3d_Nt%3d_Nx%2d_laplace_approx' % (gamma,alpha,smax,tmax,Ns,Nt,Nx)

np.save(file_name,f_rec)
        
plt.close(2)
plt.figure(2)
imgplot=plt.imshow(np.rot90(f_rec,1),origin='lower', extent=(xmin, xmax, ymax, ymin),cmap=colormap,interpolation='nearest')
plt.title(r'$\alpha = $ %4.2f, $\gamma = $ %4.2f, $s_\min = $ %4.2f, $t_\max = $ %4.2f' % (alpha,gamma,smin,tmax))
tau = np.linspace(0, 2*pi, 200)
plt.plot(p1+r1*np.cos(tau), q1+r1*np.sin(tau), color= linecolor, ls= 'dashed')
plt.plot(p2+r2*np.cos(tau), q2+r2*np.sin(tau), color= linecolor, ls= 'dashed')
plt.plot(p3+r3*np.cos(tau), q3+r3*np.sin(tau), color= linecolor, ls= 'dashed')
plt.plot([xmin,xmax],[6.5,6.5],color=linecolor,ls='dashed')

plt.colorbar()

if uniform_noise:
   file_name ='Balls3_floor_gam%4.2f_alpha%4.2f_s%4.2f_t%4.2f_Ns%3d_Nt%3d_Nx%2d_uninoise%4.2f.eps' % (gamma,alpha,smax,tmax,Ns,Nt,Nx,delta)
elif normal_noise:
   file_name ='Balls3_floor_gam%4.2f_alpha%4.2f_s%4.2f_t%4.2f_Ns%3d_Nt%3d_Nx%2d_normalnoise%4.2f.eps' % (gamma,alpha,smax,tmax,Ns,Nt,Nx,delta)
else:
    file_name ='Balls3_floor_gam%4.2f_alpha%4.2f_s%4.2f_t%4.2f_Ns%3d_Nt%3d_Nx%2d_laplace_approx.eps' % (gamma,alpha,smax,tmax,Ns,Nt,Nx)
    
plt.savefig(file_name)

if visualization_recon:
    plt.show()
