#!/usr/bin/env python
#########################################
# This module provides the function and all it subroutines to compute 
# the 2D elliptic Randon transform F\chi(s,t) of the characteristic function \chi of a ball:
#    calculate_F_Ball_2D(s,t,alpha,p1,p2,r)
#    alpha:  common offset (s+alpha,0) and (s-alpha,0) are the foci of the ellipse
#        t:  diameter of ellipse, t>2*alpha (major axis: t/2, minor axis: sqrt{t^2/4-alpha^2})
#    p1,p2,r: (p1,p2) is the center of the ball  with radius r>0
#    
# Further, the reconsruction kernel $\psi_{(p1,p2),\gamma,3}=F_approx(\Delta e_{(p1,p2),r,3})$
# is also contained:
#    Kernel_e3_2D_Laplace(s,t,alpha,p1,p2,r,Tmin,Tmax)
#    alpha:  common offset (s+alpha,0) and (s-alpha,0) are the foci of the ellipse
#        t:  diameter of ellipse, t>2*alpha (major axis: t/2, minor axis: sqrt{t^2/4-alpha^2})
#    p1,p2,r: (p1,p2) is the center of the mollifier  with radius r>0
#    Tmin,Tmax: minimal and maximal diameters such that ellipses with foci (s-alpha,0) and (s+alpha,0) 
#    and diameters between them will intersect a ball about (p1,p2) with radius r.
#
# Reference:
#     Grathwohl, Kunstmann, Quinto, Rieder:
#     Approximate inverse for the common offset acquisition geometry in 2D seismic imaging
#     http://www.waves.kit.edu/downloads/CRC1173_Preprint_2016-37.pdf
#########################################
import scipy.optimize as sop
from scipy import special
from math import *
import numpy as np


def psi(t, p1, p2, r):
    return sqrt((p1+r*cos(t))**2 + (p2+r*sin(t))**2)

def dpsi(t, p1, p2, r):
    rcost = r*cos(t)
    rsint = r*sin(t)

    return (p2*rcost-p1*rsint)/sqrt((p1+rcost)**2 + (p2+rsint)**2)

def ddpsi(t, p1, p2, r):
    rcost = r * cos(t)
    rsint = r * sin(t)

    p1_rcost = p1*rcost
    p2_rsint = p2*rsint

    denom = sqrt(((p1+rcost)**2 + (p2+rsint)**2)**3)

    return -((p1_rcost+p2_rsint+r*r)*(p1*p1+p2*p2+p1_rcost+p2_rsint))/denom

def phi(t, p1, p2, s, alpha, r):
    return psi(t, p1-s-alpha, p2 , r)+ psi(t, p1-s+alpha, p2 , r)

def dphi(t, p1, p2, s, alpha, r):
    return dpsi(t, p1-s-alpha, p2 , r)+ dpsi(t, p1-s+alpha, p2 , r)

def ddphi(t, p1, p2, s, alpha, r):
    return ddpsi(t, p1-s-alpha, p2 , r)+ ddpsi(t, p1-s+alpha, p2 , r)

def calculate_T_min_max(p1, p2, s, alpha, r):
    #################
    # This function calculates the minimal and maximal diameters such that ellipses
    # with foci (s-alpha,0) and (s+alpha,0) and diameters between them will intersect 
    # a ball about (p1,p2) with radius r. 
    #################
    
    seed = pi
    k = 2.

    t_max = sop.newton(dphi, seed/k, ddphi, args=(p1, p2, s, alpha, r))
    t_min = sop.newton(dphi, seed/k + 1.1*pi, ddphi, args=(p1, p2, s, alpha, r))
    
    phi_min = phi(t_min,p1, p2, s, alpha, r)
    phi_max = phi(t_max,p1, p2, s, alpha, r)

    
    while abs(phi_max-phi_min)<0.0001 and k<=5:
        k = k+1.
        seed = seed/k
        t_max = sop.newton(dphi, seed, ddphi, args=(p1, p2, s, alpha, r))
        t_min = sop.newton(dphi, seed + 1.1*pi, ddphi, args=(p1, p2, s, alpha, r))
        phi_min = phi(t_min,p1, p2, s, alpha, r)
        phi_max = phi(t_max,p1, p2, s, alpha, r)
        
    if abs(phi_max-phi_min)<0.0001:
        print 'ERROR in calculate_T_min_max:'
        print 'Tmax=Tmin'
        print p1, p2, s, alpha, r
        return
    elif phi_min > phi_max:
        dummy = phi_max 
        phi_max = phi_min
        phi_min = dummy
        
        dummy = t_max
        t_max = t_min
        t_min = dummy
        
        return phi_min, phi_max, t_min, t_max
    else:
        return phi_min, phi_max, t_min, t_max
    
def calculate_angles(t,p1, p2, s, alpha, r):
    ###############
    # This function computes minimal and maximal angles 
    #    min/max{\phi : \mathbf{x}(s,t,\phi)\in B_r(p1,p2)}
    # where \mathbf{x}(s,t,\phi) denotes prolate spheriodal coordinates
    #   x_1=  s+\frac{t}{2} \cos\phi
    #   x_2 = sqrt{\frac{t^2}{4}-\alpha^2}\,\sin\phi.
    ################
    b = -(p1-s)*t;
    c =  (p1-s)**2+p2**2 +(t**2)/4-alpha**2-r**2
    d = -sqrt(t**2-4*alpha**2)*p2
    
    def f(z,a,b,c,d):
        return c+b*z+(a**2)*z**2+d*np.sqrt(1-z**2)

    def df(z,a,b,c,d):
        return b+2*(a**2)*z-d*z/np.sqrt(1-z**2)
        
    def ddf(z,a,b,c,d):
        return 2*(a**2)-d/np.sqrt((1-z**2)**3)
    
    root1 = sop.newton(f, -0.9999, df, args=(alpha, b, c, d),tol = 1.e-6) # fprime2 = ddf)
    root2 = sop.newton(f, 0.9999, df, args=(alpha, b, c, d), tol = 1.e-6)#  fprime2 = ddf)
    
    phi1 = acos(root2)
    phi2 = acos(root1)

    return phi1, phi2
    
def calculate_F_Ball_2D(s,t,alpha,p1,p2,r):
    ###########################
    # The 2D elliptic Randon transform F\chi(s,t) of the characteristic function \chi of a ball:
    #    calculate_F_Ball_2D(s,t,alpha,p1,p2,r)
    #    alpha:  common offset (s+alpha,0) and (s-alpha,0) are the foci of the ellipse
    #    p1,p2,r: (p1,p2) is the center of the ball  with radius r>0
    ###########################
    
    (Tmin,Tmax,tmin,tmax)=calculate_T_min_max(p1, p2, s, alpha, r)
    
    if t < 2*alpha:
        print 'ERROR in calculate_F_Ball_2D:'
        print 'elliptic Radon transform not defined for t'
        return
    elif t >= Tmax or t <= Tmin:
        return 0
    else:
        t2=t*t
        a2=alpha*alpha
        m = a2/(t2/4-a2)
        
        (phi1,phi2)=calculate_angles(t,p1, p2, s, alpha, r)
        
        Int=np.sqrt(t2/4-a2)*special.ellipeinc(np.array([phi1,phi2]),np.array([-m,-m]))
        
        return (Int[1]-Int[0])/np.sqrt(t2-4.*a2)

        
def calculate_F_floor_2D(s,t,alpha,l):
    ###########################
    # The 2D elliptic Randon transform F\chi(s,t) of the characteristic function \chi of a half-pace x_2>=l
    #    calculate_F_Ball_2D(s,t,alpha,l)
    #    alpha:  common offset (s+alpha,0) and (s-alpha,0) are the foci of the ellipse
    #    l >0: x_2-coordinate where halfspace starts
    ###########################
    
    alpha2=alpha**2
    t2=t*t
    
    Tmin = 2*np.sqrt(l**2+alpha2)
    
    if t < 2*alpha:
        print 'ERROR in calculate_F_Ball_2D:'
        print 'elliptic Radon transform not defined for t'
        return
    elif t <= Tmin:
        return 0
    else:
        sq = np.sqrt(t2/4.-alpha2)
        phi1 = asin(l/sq)
        phi2 = pi-phi1
        m = alpha2/(t2/4-alpha2)
        
        Int=sq*special.ellipeinc(np.array([phi1,phi2]),np.array([-m,-m]))
        
        return (Int[1]-Int[0])/np.sqrt(t2-4.*alpha2)


def uu(xi):
    if xi>0:
       return exp(-1./xi)
    else:
       return 0.
    
def pp(r,M,d):
    return uu(M+d-r)/(uu(r-M-0.5*d)+uu(M+d-r))
    
def qq(r,delta):
    return uu(r/delta-1)/(uu(r/delta-1)+uu(2-r/delta))

def cut_off_t(r,cutoff,end,delta):
    if cutoff<=0:
        print 'FEHLER: cutoff muss positiv sei'
        return
    elif r>end or r<delta:
        return 0.
    elif r < 2*delta:
        return qq(r,delta)
    if abs(r)>=cutoff:
        return pp(r,cutoff,end-cutoff)
    else:
        return 1.

def cut_off_s(s,cutoff,end):
    if abs(s)<cutoff:
        return 1.
    elif abs(s)>end:
        return 0
    else:
        return pp(abs(s),cutoff,end-cutoff)

def AntiDerivative2(t1,t2,a,b,c,d):
    # anti-derivative of ((a+b*cos(t))**2 + (c*sin(t)+d)**2)**2 at t1 and t2
    # this function needs to be optimized for fast evaluation
    

    a2 = a**2
    b2 = b**2
    b4 = b2**2
    c2 = c**2
    b2_minus_c2 = b2-c2
    c4 = c2**2
    d2 = d**2
    cd = c*d
    ab = a*b
    
    P1 = (4*(a2+d2)+3*b2+c2)*ab
    P2 = b2*d2+3*a2*b2-3*c2*d2-a2*c2
    P3 = a2**2+3*c2*d2+a2*c2+d2*d2+b2*d2+2*a2*d2+3*a2*b2+0.375*(b4+c4)+b2*c2/4
    P4 = -2*ab*cd
    P5 = -(3*c2+4*(d2+a2)+b2)*cd
    P6 = ab/3
    P7 = (b2+c2)/4
    P8 = b2_minus_c2/32.
    P9 = -cd/3
    
    t = t1
    t_2 = 2*t
    t_3 = 3*t
    sint = sin(t)
    cost = cos(t)
    
    T1 = (P1+P2*cost)*sint + P3*t \
        +b2_minus_c2*(P6*sin(t_3)+P7*sin(t_2)+P8*sin(4*t)+P9*cos(t_3)) \
        +P4*cos(t_2)+P5*cost
    
    t = t2
    t_2 = 2*t
    t_3 = 3*t
    sint = sin(t)
    cost = cos(t)
    
    T2 = (P1+P2*cost)*sint + P3*t \
        +b2_minus_c2*(P6*sin(t_3)+P7*sin(t_2)+P8*sin(4*t)+P9*cos(t_3)) \
        +P4*cos(t_2)+P5*cost
    
        
    return T1,T2
        
def AntiDerivative1(t1,t2,a,b,c,d):  
    # anti-derivative of (a+b*cos(t))**2 + (c*sin(t)+d)**2 at t1 and t2
    # this function needs to be optimized for fast evaluation   
    
    b2 = b**2
    c2 = c**2
    
    P1 = a**2+(b2+c2)/2+d**2
    P2 = 2*a*b
    P3 = (b2-c2)/2
    P4 = -2*c*d
    
    t=t1
    sint = sin(t)
    
    T1 = t*P1+P2*sint+cos(t)*(P3*sint+P4)
    
    t=t2
    sint = sin(t)
    
    T2 = t*P1+P2*sint+cos(t)*(P3*sint+P4)
    
    return T1,T2
        
def Kernel_e3_2D_Laplace_approx(s,t,alpha,p1,p2,r,Tmin,Tmax):
    # This is the kernel F_approx(\Delta e_{(p1,p2),r,3})
    
    if t < 2*alpha:
        print 'ERROR in calculate_F_e1_2D:'
        print 'elliptic Radon transform not defined for t'
        return    
    elif t >= Tmax or t <= Tmin:
        return 0
    else: 
        (phi1,phi2)=calculate_angles(t,p1, p2, s, alpha, r)
        
        a = s-p1
        b = t/2
        c = sqrt(t**2/4-alpha**2)
        d = -p2
        
        (anti2,anti1) = AntiDerivative2(phi2,phi1,a,b,c,d)
        
        part2 = 36*(anti2-anti1)+12*r**4*(phi2-phi1)
        
        (anti2,anti1) = AntiDerivative1(phi2,phi1,a,b,c,d)
        
        part1 = 48*r**2*(anti2-anti1)
       
        return (part1-part2)*2./pi/r**8
